#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__    = "Tomas Fabian"
__copyright__ = "(c)2021 VSB-TUO, FEECS, Dept. of Computer Science"
__email__     = "tomas.fabian@vsb.cz"
__version__   = "0.2.0"

import math


class Point():
  """A class representing a 2D point."""
  def __init__(self, x, y):
    self.x = x
    self.y = y  
              
  def __sub__(self, other):
    """Return the difference of two points. Note that the difference of two points is a vector and not a point."""
    return Vector(self.x - other.x, self.y - other.y)
    
  def __add__(self, other):
    """Return a point translated by a vector. Make sure that the parameter other is a vector otherwise raise TypeError exception."""
    if type(other) is Vector:
      return Point(self.x + other.x, self.y + other.y)
    raise TypeException    
    
  def __str__(self):
    """String representation of a point, e.g. [1.467, 9.137]."""
    return f"[{self.x:.3f}, {self.y:.3f}]"

    
class Vector(Point):
  """A class representing a 2D vector."""
  def __init__(self, x=0.0, y=0.0):
    super().__init__(x, y)
    
  def length(self):
    """Return the length of a vector."""
    return math.sqrt(self.x**2 + self.y**2)
    
  def normalize(self):
    """Normalize the vector."""
    length = self.length()
    if length > 0:
      tmp = 1 / length
      self.x *= tmp;
      self.y *= tmp;
      
  def __sub__(self, other):
    """Return the difference of two vectors."""
    # Think about the difference of two vectors. Do we need an implementation of vector-specialized version here? The answer is (almost) no.
    if isinstance(other, (Vector,)):
      return super().__sub__(other)
    raise TypeException    

  def __neg__(self):
    """Return the opposite vector."""
    return Vector(-self.x, -self.y)
  
  def __add__(self, other):
    """Return the sum of two vectors."""
    # And what about the sum of two vectors? Can we reuse the implementation from the base class? The answer is no.
    # Note that Python preferred duck typing won't work here.
    if isinstance(other, Vector):
      return Vector(self.x + other.x, self.y + other.y)
    raise TypeException 
      
  def __mul__(self, scalar):
    """Return a vector multiplied by a scalar value."""
    return Vector(self.x * scalar, self.y * scalar)
    
  def dot(self, other):
    """Return dot product of two vectors."""
    return self.x * other.x + self.y * other.y     

  def __str__(self):
    """String representation of a vector, e.g. (9.235, -41.147)."""
    return f"({self.x:.3f}, {self.y:.3f})"


class Line(object):  
  """A class representing a line segment given by two points."""
  def __init__(self, point_a, point_b, material=None):
    self.a = point_a
    self.b = point_b
    self.material = material
    
  def direction(self):
    """Get the direction vector of this line (vector from the point A to the point B)."""
    return self.b - self.a        
    
  def normal(self, normalize=False):
    """Get the normal vector of this line. If normalize is True then the normal vector should be of lenght 1."""
    normal = self.b - self.a
    normal.x, normal.y = -normal.y, normal.x
    if normalize:
      normal.normalize()
    return normal
  
  def mid_point(self):
    """Mid point of the line."""
    return Point((self.a.x + self.b.x) / 2, (self.a.y + self.b.y) / 2)

  def intersection(self, other, on_first=True, on_second=True):
    """Find the intersection point (if exists) of two lines."""
    d1 = self.direction()
    d2 = other.direction()
    det = d1.y * d2.x - d1.x * d2.y
    if det != 0:
      da = other.a - self.a      
      t1 = (d2.x * da.y - d2.y * da.x) / det
      t2 = (d1.x * da.y - d1.y * da.x) / det
      inside1 = t1 > 0 and t1 < 1
      inside2 = t2 > 0 and t2 < 1
      if (not on_first or (on_first and inside1)) and (not on_second or (on_second and inside2)):
        i1 = self.a + d1 * t1
        i2 = other.a + d2 * t2
        return i1
    return None            
      
  def split_by(self, other):
    """Split this line by the other one and get the position of the result(s) relatively to the normal of the other line."""
    i = self.intersection(other, True, False)

    if i is None:    
      front_line = self
      back_line = None      
    else:
      front_line = Line(self.a, i, self.material)
      back_line = Line(i, self.b, self.material)

    # is "front line" really in front of the other one?    
    if (front_line.mid_point() - other.a).dot(other.normal()) < 0:
      front_line, back_line = back_line, front_line              
      
    return (front_line, back_line)    

  def affine_transform(self, pre_translation, rotation, scale=1, post_translation=Vector(0, 0)):
    """Return a new transformed line."""
    # pre translation
    a = self.a + pre_translation
    b = self.b + pre_translation
    # rotation
    cos = math.cos(rotation)
    sin = math.sin(rotation)
    a = Point(a.x * cos - a.y * sin, a.x * sin + a.y * cos)
    b = Point(b.x * cos - b.y * sin, b.x * sin + b.y * cos)
    # scale
    a.x *= scale
    a.y *= scale
    b.x *= scale
    b.y *= scale
    # post translation
    a += post_translation
    b += post_translation
    
    return Line(a, b, self.material)   
    
  def __str__(self):
    """String representation of a line, e.g. [[1.267, 9.479] -> [15.267, 49.457]."""
    return f"[{self.a} -> {self.b}]"


def main():
  a = Point(2, 3)
  print(a)
  
  b = Point(9, 4)
  d = b - a
  print(d)
  print(d * 3)
  print(d.length())
  d.normalize()
  print(d.length())  
  
  line = Line(a, b)
  print(line)
  normal = line.normal()  
  print(normal)
  print(normal.dot(d))
    
  line1 = Line(Point(1,1), Point(4,4))
  
  print(line1.normal())
  line2 = Line(Point(2,4), Point(2,5))
  print("line2 normal is " + str(line2.normal()))  
  print(line1.intersection(line2))  
   
  print("split " + str(line1) + " by " + str(line2))  
  front_line, back_line = line1.split_by(line2)
  print(str(front_line) + " is in front")  
  print(str(back_line) + " is in back")


if __name__ == "__main__":
  main()
